"""Common classes and methods for all message types."""
import logging
from abc import ABC
from datetime import datetime
from enum import Enum, unique
from typing import Any, ClassVar, Dict, Generic, Optional, Sequence, TypeVar, Union, cast
from uuid import UUID

from fred_types import BaseModel, GenericModel
from typist import SecretaryClient
from typist.client import TemplateNotFound

from .utils import AbcAttribute

try:
    from pydantic.v1 import Field
except ImportError:  # pragma: no cover
    from pydantic import Field  # type: ignore[no-redef, assignment]

LOGGER = logging.getLogger(__name__)


@unique
class MessageStatus(int, Enum):
    """Message status."""

    PENDING = 0
    SENT = 1
    FAILED = 2
    ERROR = 3
    DELIVERED = 4
    UNDELIVERED = 5
    CANCELED = 6


class Reference(BaseModel):
    """Registry object reference."""

    type: str
    value: str


class Message(BaseModel, ABC):
    """Base class for all message types."""

    type: Optional[str] = None
    _secretary_client: Optional[SecretaryClient] = None
    _secretary_property_cache: Dict[str, str] = {}

    render_fields: ClassVar[Sequence[str]] = AbcAttribute

    class Config:
        underscore_attrs_are_private = True

    def dict(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """Return dictionary with message info.

        This method is useful for message serialization.
        """
        d = super().dict(*args, **kwargs)
        if self._secretary_client:
            d.update(self._get_secretary_client_data())
        return cast(Dict[str, Any], d)  # type: ignore[redundant-cast]

    def _secretary_render(
        self, name: str, *, template: Union[UUID, str], render_method_name: str = "render"
    ) -> str:
        """Render property by django-secretary method.

        These properties are cached. The template and context attributes
        are not supposed to be overwritten.
        If you change them, you have to invalidate the cache manually.

        Args:
            name: Name of rendered property. Used for caching the results.
            template: UUID or name of template to be rendered.
            render_method_name: Name of secretary render method.

        Returns:
            Rendered template.

        Raises:
            RuntimeError: If secretary client is not set.
        """
        if name not in self._secretary_property_cache:
            if not self._secretary_client:
                raise RuntimeError(
                    "Can't render {}, because secretary_client is not set".format(name)
                )

            self._secretary_property_cache[name] = getattr(
                self._secretary_client, render_method_name
            )(template, context=getattr(self, "context", {}))
        return self._secretary_property_cache[name]

    def _get_secretary_client_data(self) -> Dict[str, Any]:
        """Return dictionary with SecretaryClient field values and render errors.

        When getting SecretaryClient data, which are being rendered, a TemplateNotFound
        error can occur. In this case None is provided as the value and a record is added
        into the returned dictionary in the ``render_errors`` list.
        """
        ret: Dict[str, Any] = {}
        errors = []

        for item in self.render_fields:
            try:
                ret[item] = getattr(self, item)
            except (RuntimeError, TemplateNotFound, ValueError) as e:
                ret[item] = None
                errors.append(
                    'SecretaryClient raised an error in the "{}" property: {}'.format(item, e)
                )
                LOGGER.warning(
                    'SecretaryClient raised an error in the "{}" property: {}'.format(item, e)
                )

        if errors:
            ret["render_errors"] = errors

        return ret


M = TypeVar("M", bound=Message)


class MessageEnvelope(GenericModel, ABC, Generic[M]):
    """Base class for all message envelope types."""

    message: M = Field(alias="message_data")
    uid: Optional[str] = None
    create_datetime: Optional[datetime] = None
    send_datetime: Optional[datetime] = None
    delivery_datetime: Optional[datetime] = None
    attempts: int = 0
    status: Optional[MessageStatus] = None
    archive: bool = True
    references: Sequence[Reference] = []

    class Config:
        allow_population_by_field_name = True
        populate_by_name = True


E = TypeVar("E", bound=MessageEnvelope)
