"""Unittests for generic message structures."""
from typing import cast
from unittest import TestCase
from unittest.mock import Mock, call, patch, sentinel

from typist import SecretaryClient
from typist.client import TemplateNotFound

from hermes.message import Message, MessageEnvelope


class TestMessage(Message):
    """Test message."""

    data: str
    render_fields = ("subject",)

    @property
    def subject(self) -> str:
        return self._secretary_render("subject", template="template.txt")


class TestMessageEnvelope(MessageEnvelope):
    """Test message envelope."""

    message_cls = TestMessage


class MessageTest(TestCase):
    """Unittests for message."""

    def test_dict(self):
        message = TestMessage(data="gazpacho", type="T")
        self.assertEqual(message.dict(), {"data": "gazpacho", "type": "T"})

    def test_secretary_render(self):
        """Test _secretary_render without secretary client."""
        message = TestMessage(data="gazpacho")
        with self.assertRaises(RuntimeError):
            message._secretary_render(
                sentinel.name,
                template=sentinel.template,
                render_method_name=sentinel.render_method_name,
            )


class MessageWithSecretaryClientTest(TestCase):
    """Unittests for message structure with a secretary client."""

    def setUp(self):
        self.secretary_client = SecretaryClient("localhost")
        patcher = patch.object(self.secretary_client, "render")
        self.addCleanup(patcher.stop)
        patcher.start()
        self.message = TestMessage(data="gazpacho", type="T")
        self.message._secretary_client = self.secretary_client

    def test_secretary_render(self):
        """Test _secretary_render with secretary client."""
        cast(Mock, self.secretary_client).render.return_value = sentinel.subject
        self.assertEqual(self.message.subject, sentinel.subject)
        self.assertEqual(
            self.message.subject, sentinel.subject
        )  # call subject again to test the caching
        self.assertEqual(
            cast(Mock, self.secretary_client).render.mock_calls, [call("template.txt", context={})]
        )

    def test_dict(self):
        cast(Mock, self.secretary_client).render.return_value = sentinel.subject
        self.assertEqual(
            self.message.dict(),
            {
                "data": "gazpacho",
                "type": "T",
                "subject": sentinel.subject,
            },
        )

    def test_get_secretary_client_data_with_errors(self):
        for error_cls in [RuntimeError, TemplateNotFound, ValueError]:
            with self.subTest(error_cls=error_cls):
                cast(Mock, self.secretary_client).render.side_effect = error_cls("Dangerous error")
                with self.assertLogs("hermes.message", level="WARNING") as log:
                    data = self.message._get_secretary_client_data()
                self.assertEqual(
                    data,
                    {
                        "subject": None,
                        "render_errors": [
                            'SecretaryClient raised an error in the "subject" property: '
                            "Dangerous error",
                        ],
                    },
                )
                self.assertEqual(
                    log.output,
                    [  # type: ignore[union-attr]
                        "WARNING:hermes.message:SecretaryClient raised an error "
                        'in the "subject" property: Dangerous error',
                    ],
                )

    def test_secretary_client_fields_error_raised(self):
        cast(Mock, self.secretary_client).render.side_effect = TemplateNotFound
        # call twice to test nothing has been cached
        with self.assertRaises(TemplateNotFound):
            self.message.subject
        with self.assertRaises(TemplateNotFound):
            self.message.subject
        self.assertEqual(
            cast(Mock, self.secretary_client).render.mock_calls,
            [call("template.txt", context={}), call("template.txt", context={})],
        )
